/*
  The contents of this file are subject to the University of Utah Public
  License (the "License"); you may not use this file except in
  compliance with the License.
  
  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Source Code is "teem", released March 23, 2001.
  
  The Original Source Code was developed by the University of Utah.
  Portions created by UNIVERSITY are Copyright (C) 2001, 1998 University
  of Utah. All Rights Reserved.
*/


#include "air.h"

/*
******** airSanity()
**
** Does run-time checks to see if the compile-time constants are correct.
** Returns a value 
*/
int
airSanity(void) {
  float nan, inf;
  int tmpI, sign, exp, frac;
  char endian;
  static int sanity=0;
  
  if (sanity) {
    return airInsane_not;
  }

  /* run-time endian check */
  tmpI = 1;
  endian = !(*((char*)(&tmpI)));
  if (endian) {
    /* big endian */
    if (4321 != airMyEndian) {
      return airInsane_endian;
    }
  }
  else {
    if (1234 != airMyEndian) {
      return airInsane_endian;
    }
  }    

  /* run-time NaN checks */
  inf = 3e+38F;                 /* close to FLT_MAX */
  inf = inf * inf * inf * inf;  /* this generates infinity even for doubles */
  inf = inf * inf * inf * inf;
  inf = inf * inf * inf * inf;
  inf = inf * inf * inf * inf;
  if (AIR_EXISTS(inf)) {
    return airInsane_pInfExists;
  }
  inf = -inf;
  if (AIR_EXISTS(inf)) {
    return airInsane_nInfExists;
  }
  nan = inf/inf;
  if (AIR_EXISTS(nan)) {
    return airInsane_NaNExists;
  }
  airFPValToParts(&sign, &exp, &frac, nan);
  frac >>= 22;
  if (airMyQNaNHiBit != frac) {
    return airInsane_QNaNHiBit;
  }
  
  /* just make sure airMyDio is reasonably set */
  switch (airMyDio) {
  case 0: break;
  case 1: break;
  default:
    return airInsane_dio;
  }

  sanity = 1;
  return airInsane_not;
}

const char
_airInsaneErr[8][AIR_STRLEN_SMALL] = {
  "(invalid insane value)",
  "sanity checked PASSED!",
  "airMyEndian is wrong",
  "AIR_EXISTS(+inf) was true",
  "AIR_EXISTS(-inf) was true",
  "AIR_EXISTS(NaN) was true",
  "airMyQNaNHiBit is wrong",
  "airMyDio has invalid value"
};

const char *
airInsaneErr(int insane) {
  
  if (AIR_INSIDE(0, insane, 6)) {
    return _airInsaneErr[insane+1];
  }
  else {
    return _airInsaneErr[0];
  }
}

/*
******** airNull()
**
** returns NULL
*/
void *
airNull(void) {

  return NULL;
}

/*
******** airSetNull
**
** dereferences and sets to NULL, returns NULL
*/
void *
airSetNull(void **ptrP) {
  
  if (ptrP) {
    *ptrP = NULL;
  }
  return NULL;
}

/*
******** airFree()
**
** to facilitate setting a newly free()'d pointer; always returns NULL.
** also makes sure that NULL is not passed to free().
*/
void *
airFree(void *ptr) {

  if (ptr) {
    free(ptr);
  }
  return NULL;
}

/*
******** airFreeP()
**
** calls airFree on the address pointed to by the argument
*/
void *
airFreeP(void *_ptrP) {
  void **ptrP;

  ptrP = _ptrP;
  if (ptrP)
    airFree(*ptrP);
  return NULL;
}

/*
******** airFclose()
**
** just to facilitate setting a newly fclose()'d file pointer to NULL
** also makes sure that NULL is not passed to fclose()
*/
FILE *
airFclose(FILE *file) {

  if (file)
    fclose(file);
  return NULL;
}

/*
******** airStrdup()
**
** because they didn't put strdup() in ANSI ...
** This will return NULL if given NULL.
*/
char *
airStrdup(const char *s) {
  char *ret;

  if (!s) {
    ret = NULL;
  }
  else {
    ret = malloc(strlen(s)+1);
    if (ret) {
      strcpy(ret, s);
    }
  }
  return ret;
}

/*
******** airStrlen()
**
** just like strlen, but safe to call on NULL (for which return is 0)
*/
size_t
airStrlen(const char *s) {
  size_t ret;
  
  if (!s) {
    ret = 0;
  }
  else {
    ret = strlen(s);
  }
  return ret;
}

/*
******** airStrtok()
**
** thread-safe strtok() replacement.  Use just like strtok(), but on
** each call to parse a given string, pass as the last argument the
** address of a char*, to be used for saving state while the string is
** traversed.  Like strtok(), this will alter the "s" array passed to
** it on the first call, and like strtok(), this returns pointers into
** this string (rather than allocating new strings for each token).
*/
char *
airStrtok(char *s, char *ct, char **last) {
  char *h, *e;
  
  if (!(ct && last)) {
    /* can't do any work, bail */
    return NULL;
  }
  h = s ? s : *last;
  if (!airStrlen(h))
    return NULL;
  h += strspn(h, ct);
  e = h + strcspn(h, ct);
  if ('\0' == *e) {
    *last = e;
  }
  else {
    *e = '\0';
    *last = e + 1;
  }
  return h;
}

/*
******** airStrntok()
**
** returns the number of tokens parsable by airStrtok(), but does
** NOT alter the given string
*/
int
airStrntok(char *_s, char *ct) {
  char *s, *t, *l;
  int n = 0;

  if (_s && ct) {
    s = airStrdup(_s);
    t = airStrtok(s, ct, &l);
    while (t) {
      n++;
      t = airStrtok(NULL, ct, &l);
    }
    airFree(s);
  }
  return n;
}

char *
airStrtrans(char *s, char from, char to) {
  int i, l;
  
  if (s) {
    l = strlen(s);
    for (i=0; i<=l-1; i++) {
      s[i] = (s[i] == from ? to : s[i]);
    }
  }
  return s;
}

/*
******** airEndsWith
**
** if "s" ends with "suff", then returns 1, 0 otherwise
*/
int
airEndsWith(const char *s, const char *suff) {
  
  if (!(s && suff))
    return 0;
  if (!(strlen(s) >= strlen(suff)))
    return 0;
  if (!strncmp(s + strlen(s) - strlen(suff), suff, strlen(suff)))
    return 1;
  else
    return 0;
}

/*
******** airOneLinify()
**
** converts all contiguous white space (as determined by isspace()) to
** a single ' ', entirely removes non-printable (as determined by
** isprint()) characters, and entirely removes white space contiguous
** with the end of the string, even if that means shrinking the string
** to "".
**
** Useful for cleaning up lines of text to be saved as strings in
** fields of other structs
*/
char *
airOneLinify(char *s) {
  int i, j, len;

  len = airStrlen(s);
  if (!len) 
    return s;

  /* convert white space to space (' '), and delete unprintables */
  for (i=0; i<=len-1; i++) {
    if (isspace(s[i])) {
      s[i] = ' ';
      continue;
    }
    if (!isprint(s[i])) {
      for (j=i; j<=len-1; j++) {
	/* this will copy the '\0' at the end */
	s[j] = s[j+1];
      }
      i--;
      continue;
    }
  }

  /* compress all contiguous spaces into one */
  for (i=0; i<=len-1; i++) {
    while (' ' == s[i] && ' ' == s[i+1]) {
      for (j=i+1; j<=len-1; j++) {
	s[j] = s[j+1];
      }
    }
  }

  /* lose trailing white space */
  len = airStrlen(s);
  for (i=len-1; i>=0 && ' ' == s[i]; i--)
    s[i] = '\0';

  return s;
}

/*
******* airDoneStr()
**
** dinky little utility for generating progress messages of the form
** "  1.9%" or " 35.3%" or  "100.0%"
**
** The message will ALWAYS be six characters, and will ALWAYS be
** preceeded by six backspaces.  Thus, you pass in a string to print
** into, and it had better be allocated for at least 6+6+1 = 13 chars.
*/
char *
airDoneStr(float start, float here, float end, char *str) {
  int perc=0;
  static int len = -1;

  if (str) {
    if (end != start)
      perc = (int)(1000*(here-start)/(end-start) + 0.5);
    else
      perc = 1000;
    if (perc < 0) {
      sprintf(str, "\b\b\b\b\b\b ---- ");
    } else if (perc < 1000) {
      sprintf(str, "\b\b\b\b\b\b% 3d.%d%%", perc/10, perc%10);
    }
    else if (perc == 1000) {
      /* the "% 3d" formatting sequence should have taken care
	 of this, but whatever */
      sprintf(str, "\b\b\b\b\b\b100.0%%");
    }
    else {
      sprintf(str, "\b\b\b\b\b\b  done");
    }
  }

  if (-1 == len) {
    len = strlen(str);
  }
  else {
    if (len != strlen(str)) {
      printf("len(\\b\\b\\b\\b\\b\\b% 3d.%d%%) != %d\n", 
	      perc/10, perc%10, len);
      exit(1);
    }
  }
  return(str);
}

/*
******** airTime()
**
** returns current time in milliseconds as a double
*/
double
airTime() {
  struct timeval tv;
  struct timezone tz;

  gettimeofday(&tv, &tz); 
  return((double)(tv.tv_sec + tv.tv_usec/1000000.0));
}

/*
******** airLog2()
**
** silly little function which returns log_2(n) if n is a power of 2,
** or -1 otherwise
*/
int 
airLog2(float n) {

  if (!AIR_EXISTS(n))
    return -1;
  if (n == 1.0)
    return 0;
  if (n < 2)
    return -1;
  return airLog2(n/2.0);
}

/*
******** airToLower()
**
** calls tolower() on all characters in a string, and returns the same
** pointer that it was given
*/
char *
airToLower(char *str) {
  char *c;

  if (str) {
    c = str;
    while (*c) {
      *c = tolower(*c);
      c++;
    }
  }
  return str;
}

/*
******** airToUpper()
**
** calls toupper() on all characters in a string, and returns the same
** pointer that it was given
*/
char *
airToUpper(char *str) {
  char *c;

  if (str) {
    c = str;
    while (*c) {
      *c = toupper(*c);
      c++;
    }
  }
  return str;
}

/*
******** airOneLine()
** 
** gets one line from "file", putting it into an array if given size.
** "size" must be the size of line buffer "line".  Always null-terminates
** the contents of the array (except if the arguments are invalid).
**
** -1: if arguments are invalid
** 0: if saw EOF before seeing a newline
** 1: if line was a single newline
** n, where n <= size: if line was n-1 characters followed by newline
** size+1: if didn't see a newline within size-1 characters
**
** So except for returns of -1 and size+1, the return is the number of
** characters comprising the line, including the newline character.
*/
int
airOneLine(FILE *file, char *line, int size) {
  int c=0, i;
  
  if (!(size >= 2 && line && file))
    return -1;
  line[0] = 0;
  for (i=0; (i <= size-2 && 
	     EOF != (c=getc(file)) && 
	     c != '\n'); ++i)
    line[i] = c;
  if (EOF == c)
    return 0;
  if (0 == i) {
    /* !EOF ==> we stopped because of newline */
    return 1;
  }
  if (i == size-1) {
    line[size-1] = 0;
    if ('\n' != c) {
      /* we got to the end of the array and still no newline */
      return size+1;
    }
    else {
      /* we got to end of array just as we saw a newline, all's well */
      return size;
    }
  }
  /* i < size-1 && EOF != c ==> '\n' == c */
  line[i] = 0;
  return i+1;
}

/*
******** airSinglePrintf
**
** a complete stand-in for {f|s}printf(), as long as the given format
** string contains exactly one conversion sequence.  The utility of
** this is to standardize the printing of IEEE 754 special values:
** QNAN, SNAN -> "NaN"
** POS_INF -> "+inf"
** NEG_INF -> "-inf"
** The format string can contain other things besides just the
** conversion sequence: airSingleFprintf(f, " (%f)\n", AIR_NAN)
** will be the same as fprintf(f, " (%s)\n", "NaN");
**
** To get fprintf behavior, pass "str" as NULL
** to get sprintf bahavior, pass "file" as NULL
**
** Someday I'll find/write a complete {f|s|}printf replacement ...
*/
int
airSinglePrintf(FILE *file, char *str, char *_fmt, ...) {
  char *fmt;
  float valF;
  double valD=0;
  int ret, isF, isD;
  char *conv=NULL, *p0, *p1, *p2, *p3, *p4, *p5;
  va_list ap;
  
  va_start(ap, _fmt);
  fmt = airStrdup(_fmt);
  
  p0 = strstr(fmt, "%e");
  p1 = strstr(fmt, "%f");
  p2 = strstr(fmt, "%g");
  p3 = strstr(fmt, "%le");
  p4 = strstr(fmt, "%lf");
  p5 = strstr(fmt, "%lg");
  isF = p0 || p1 || p2;
  isD = p3 || p4 || p5;
  if (isF) {
    conv = p0 ? p0 : (p1 ? p1 : p2);
  }
  if (isD) {
    conv = p3 ? p3 : (p4 ? p4 : p5);
  }
  if (isF || isD) {
    if (isF) {
      /* use "double" instead of "float" because var args are _always_
	 subject to old-style C type promotions: float promotes to double */
      valF = (float)(va_arg(ap, double));
    }
    else {
      valD = va_arg(ap, double);
      valF = valD;
    }
    switch (airFPClass_f(valF)) {
    case airFP_SNAN:
    case airFP_QNAN:
    case airFP_POS_INF:
    case airFP_NEG_INF:
      if (isF) {
	memcpy(conv, "%s", 2);
      }
      else {
	/* this sneakiness allows us to replace a 3-character conversion
	   sequence for a double (such as %lg) with a 3-character conversion
	   for a string, which we know has at most 4 characters */
	memcpy(conv, "%4s", 3);
      }
      break;
    }
#define PRINT(F, S, C, V) ((F) ? fprintf((F),(C),(V)) : sprintf((S),(C),(V)))
    switch (airFPClass_f(valF)) {
    case airFP_SNAN:
    case airFP_QNAN:
      ret = PRINT(file, str, fmt, "NaN");
      break;
    case airFP_POS_INF:
      ret = PRINT(file, str, fmt, "+inf");
      break;
    case airFP_NEG_INF:
      ret = PRINT(file, str, fmt, "-inf");
      break;
    default:
      if (isF) {
	ret = PRINT(file, str, fmt, valF);
      }
      else {
	ret = PRINT(file, str, fmt, valD);
      }
      break;
    }
  }
  else {
    ret = file ? vfprintf(file, fmt, ap) : vsprintf(str, fmt, ap);
  }
  
  va_end(ap);
  free(fmt);
  return ret;
}

void
airBinaryPrintUInt(FILE *file, int digits, unsigned int N) {

  digits = AIR_CLAMP(1, digits, 32);
  for (digits=digits; digits>=1; digits--) {
    fprintf(file, "%c", ((1<<(digits-1)) & N
			 ? '1' : '0'));
  }
}

double
airErf(double x) {
  /* 
   * When I was a Cornell undergrad (sophomore year, 1992), I was
   * interested in a programming job in the Astronomy department.  The
   * job posting said I should talk to Saul Teukolsky, one of the
   * Numerical Recipes authors.  The first thing he asked was "What's
   * your GPA?".  I was confused.  If programming ability is the real
   * question at hand, this was about as sensible as asking "How much
   * do you weigh?"  As soon as I admitted to getting a B+ in a
   * previous physics class, he simply waved me away and turned back
   * to his computer.
   */
  double t,z,ans;

  z = AIR_ABS(x);
  t = 1.0/(1.0+0.5*z);
  ans=t*exp(-z*z-1.26551223+t*(1.00002368+t*(0.37409196+t*(0.09678418+
            t*(-0.18628806+t*(0.27886807+t*(-1.13520398+t*(1.48851587+
            t*(-0.82215223+t*0.17087277)))))))));
  return (x >= 0.0 ? 1.0 - ans : ans - 1.0);
}

const char
airTypeStr[AIR_TYPE_MAX+1][AIR_STRLEN_SMALL] = {
  "(unknown)",
  "bool",
  "int",
  "float",
  "double",
  "char",
  "string"
};

const int
airTypeSize[AIR_TYPE_MAX+1] = {
  0,
  sizeof(int),
  sizeof(int),
  sizeof(float),
  sizeof(double),
  sizeof(char),
  sizeof(char*),
  0   /* we don't know anything about type "other" */
};

int
airILoad(void *v, int t) {

  switch(t) {
  case airTypeBool:   return *((int*)v); break;
  case airTypeInt:    return *((int*)v); break;
  case airTypeFloat:  return *((float*)v); break;
  case airTypeDouble: return *((double*)v); break;
  case airTypeChar:   return *((char*)v); break;
  default: return 0; break;
  }
}

int
airIStore(void *v, int t, int i) {

  switch(t) {
  case airTypeBool:   return (*((int*)v) = !!i); break;
  case airTypeInt:    return (*((int*)v) = i); break;
  case airTypeFloat:  return (*((float*)v) = i); break;
  case airTypeDouble: return (*((double*)v) = i); break;
  case airTypeChar:   return (*((char*)v) = i); break;
  default: return 0; break;
  }
}

float
airFLoad(void *v, int t) {

  switch(t) {
  case airTypeBool:   return *((int*)v); break;
  case airTypeInt:    return *((int*)v); break;
  case airTypeFloat:  return *((float*)v); break;
  case airTypeDouble: return *((double*)v); break;
  case airTypeChar:   return *((char*)v); break;
  default: return 0; break;
  }
}

float
airFStore(void *v, int t, float f) {

  switch(t) {
  case airTypeBool:   return (*((int*)v) = f); break;
  case airTypeInt:    return (*((int*)v) = f); break;
  case airTypeFloat:  return (*((float*)v) = f); break;
  case airTypeDouble: return (*((double*)v) = f); break;
  case airTypeChar:   return (*((char*)v) = f); break;
  default: return 0; break;
  }
}

double
airDLoad(void *v, int t) {

  switch(t) {
  case airTypeBool:   return *((int*)v); break;
  case airTypeInt:    return *((int*)v); break;
  case airTypeFloat:  return *((float*)v); break;
  case airTypeDouble: return *((double*)v); break;
  case airTypeChar:   return *((char*)v); break;
  default: return 0; break;
  }
}

double
airDStore(void *v, int t, double d) {

  switch(t) {
  case airTypeBool:   return (*((int*)v) = d); break;
  case airTypeInt:    return (*((int*)v) = d); break;
  case airTypeFloat:  return (*((float*)v) = d); break;
  case airTypeDouble: return (*((double*)v) = d); break;
  case airTypeChar:   return (*((char*)v) = d); break;
  default: return 0; break;
  }
}

