/*
  The contents of this file are subject to the University of Utah Public
  License (the "License"); you may not use this file except in
  compliance with the License.
  
  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Source Code is "teem", released March 23, 2001.
  
  The Original Source Code was developed by the University of Utah.
  Portions created by UNIVERSITY are Copyright (C) 2001, 1998 University
  of Utah. All Rights Reserved.
*/


#include "air.h"

int
airSingleSscanf(char *str, char *fmt, void *ptr) {
  char *tmp;
  double val;
  int ret;
  
  if (!strcmp(fmt, "%e") || !strcmp(fmt, "%f") || !strcmp(fmt, "%g")
      || !strcmp(fmt, "%le") || !strcmp(fmt, "%lf") || !strcmp(fmt, "%lg")) {
    tmp = airStrdup(str);
    if (!tmp) {
      return 0;
    }
    airToLower(tmp);
    if (strstr(tmp, "nan")) {
      val = AIR_NAN;
    }
    else if (strstr(tmp, "-inf")) {
      val = AIR_NEG_INF;
    }
    else if (strstr(tmp, "inf")) {
      val = AIR_POS_INF;
    }
    else {
      /* nothing special matched; pass it off to sscanf() */
      ret = sscanf(str, fmt, ptr);
      free(tmp);
      return ret;
    }
    /* else we matched "nan", "-inf", or "inf", and set val accordingly */
    if (!strncmp(fmt, "%l", 2)) {
      /* we were given a double pointer */
      *((double *)(ptr)) = val;
    }
    else {
      /* we were given a float pointer */
      *((float *)(ptr)) = val;
    }
    free(tmp);
    return 1;
  }
  else {
    /* this is neither a float nor double */
    return sscanf(str, fmt, ptr);
  }
}

#define _PARSE_STR_ARGS(type, format) type *out, char *_s, char *ct, int n
#define _PARSE_STR_BODY(type, format) \
  int i; \
  char *tmp, *s, *last; \
  \
  /* if we got NULL, there's nothing to do */ \
  if (!(out && _s && ct)) \
    return 0; \
  \
  /* copy the input so that we don't change it */ \
  s = airStrdup(_s); \
  \
  /* keep calling airStrtok() until we have everything */ \
  for (i=0; i<=n-1; i++) { \
    tmp = airStrtok(i ? NULL : s, ct, &last); \
    if (!tmp) { \
      free(s); \
      return i; \
    } \
    if (1 != airSingleSscanf(tmp, format, out+i)) { \
      free(s); \
      return i; \
    } \
  } \
  free(s); \
  return n; \

/*
******* airParse*()
**
** parse ints, floats, doubles, or single characters, from some
** given string "s"; try to parse "n" of them, as delimited by
** characters in "ct", and put the results in "out".
**
** Returns the number of things succesfully parsed- should be n; 
** there's been an error if return is < n.
**
** int airParseStrI(int *out, char *s, char *ct, int n)
** int airParseStrF(float *out, char *s, char *ct, int n)
** int airParseStrD(double *out, char *s, char *ct, int n)
** int airParseStrC(char *out, char *s, char *ct, int n)
**
** This uses air's thread-safe strtok() replacement: airStrtok()
*/
int
airParseStrI(_PARSE_STR_ARGS(int, "%d"))     { _PARSE_STR_BODY(int, "%d") }

int
airParseStrF(_PARSE_STR_ARGS(float, "%f"))   { _PARSE_STR_BODY(float, "%f") }

int
airParseStrD(_PARSE_STR_ARGS(double, "%lf")) { _PARSE_STR_BODY(double, "%lf") }

double
airAtod(char *str) {
  double val = 0.0;
  
  airSingleSscanf(str, "%lf", &val);
  return val;
}

int
airParseStrB(int *out, char *_s, char *ct, int n) {
  int i;
  char *tmp, *s, *last;

  /* if we got NULL, there's nothing to do */
  if (!(out && _s && ct))
    return 0;

  /* copy the input so that we don't change it */
  s = airStrdup(_s);

  /* keep calling airStrtok() until we have everything */
  for (i=0; i<=n-1; i++) {
    tmp = airStrtok(i ? NULL : s, ct, &last);
    if (!tmp) {
      free(s);
      return i;
    }
    airToLower(tmp);
    if (!strcmp(tmp, "on") || !strcmp(tmp, "yes") || !strcmp(tmp, "y") ||
	!strcmp(tmp, "true") ||	!strcmp(tmp, "t") ||
	!strcmp(tmp, "1") || !strcmp(tmp, "yup") ||
	!strcmp(tmp, "yea") || !strcmp(tmp, "ofcourse")) {
      out[i] = 1;
    }
    else if (!strcmp(tmp, "off") || !strcmp(tmp, "no") || !strcmp(tmp, "n") || 
	     !strcmp(tmp, "false") || !strcmp(tmp, "f") ||
	     !strcmp(tmp, "0") || !strcmp(tmp, "nope") ||
	     !strcmp(tmp, "nah") || !strcmp(tmp, "noway")) {
      out[i] = 0;
    }
    else {
      free(s);
      return i;
    }
  }
  free(s);
  return n;
}

int
airParseStrS(char **out, char *_s, char *ct, int n) {
  int i;
  char *tmp, *s, *last;
  airArray *mop;

  /* if we got NULL, there's nothing to do */
  if (!(out && _s && ct))
    return 0;

  mop = airMopInit();
  /* copy the input so that we don't change it */
  s = airStrdup(_s);
  airMopMem(mop, &s, airMopAlways);

  /* keep calling airStrtok() until we have everything */
  for (i=0; i<=n-1; i++) {
    tmp = airStrtok(i ? NULL : s, ct, &last);
    if (!tmp) {
      airMopError(mop);
      return i;
    }
    out[i] = airStrdup(tmp);
    if (!out[i]) {
      airMopError(mop);
      return i;
    }
    airMopMem(mop, out+i, airMopOnError);
  }
  airMopOkay(mop);
  return n;
}

int
airParseStrC(char *out, char *_s, char *ct, int n) {
  int i;
  char *tmp, *s, *last;

  /* if we got NULL, there's nothing to do */
  if (!(out && _s && ct))
    return 0;

  /* copy the input so that we don't change it */
  s = airStrdup(_s);

  /* keep calling airStrtok() until we have everything */
  for (i=0; i<=n-1; i++) {
    tmp = airStrtok(i ? NULL : s, ct, &last);
    if (!tmp) {
      free(s);
      return i;
    }
    out[i] = tmp[0];
  }
  free(s);
  return n;
}

int
(*airParseStr[AIR_TYPE_MAX+1])(void *, char *, char *, int) = {
  NULL,
  (int (*)(void *, char *, char *, int))airParseStrB,
  (int (*)(void *, char *, char *, int))airParseStrI,
  (int (*)(void *, char *, char *, int))airParseStrF,
  (int (*)(void *, char *, char *, int))airParseStrD,
  (int (*)(void *, char *, char *, int))airParseStrC,
  (int (*)(void *, char *, char *, int))airParseStrS,
  NULL   /* no standard way of parsing type "other" */
};

